/*
kingpairs.c 0.3.0 (2003-Apr-10-Thu)
Adam M. Costello
http://www.nicemice.net/amc/

kingpairs [ -c <c> ] [ -s <s> ]

Reads an even number (call it 2n) of whitespace-separated words from
stdin describing n DNS servers, measures up to n*(n-1) round-trip times
(one for each pair of servers), using between 2*c+1 and 3*c probes each,
and writes the results to stdout.  The default value of c is 4.  Each
server is allowed s "strikes" before it is "out" (pruned, see below).
The default value of s is 1.

Input format:

The input words must be in pairs: an IP address (in dotted decimal
notation--currently only IPv4 is supported) followed by a domain
name (in simple dot-delimited text form, with no quotation or escape
mechanisms).

The IP address should belong to a DNS server that allows recursive
queries, but violations of this expectation will be detected.

The domain name is assumed to be a domain for which the DNS server is
authoritative (typically it ends with .in-addr.arpa).  Furthermore, it
is assumed that *all* authoritative DNS servers for the domain are in
the same place as the specified server (or close enough to ignore the
difference).  Violations of these assumptions will *not* be detected,
so it is the caller's responsibility to check these assumptions before
running kingpairs.  An approximate check for the latter assumption is
to check that all IP addresses of all authoritative DNS servers for the
domain share a 24-bit prefix.

If the IP address is "pair" and the domain name is "list", that is a
special marker that terminates the server list and begins a task list.
The remaining pairs are ordered pairs <id1> <id2> used to initialize
the list of round-trip measurements to attempt (see below for the
exact syntax and meaning of <id1> and <id2>).  If the server list is
terminated by EOF rather than the special marker, the task list is
initialized to contain every ordered pair of distinct server IDs.

Output format:

The first four output lines are:

units_per_second <u>
redundancy <c>
strikes_per_out <s>
nodes <n>

All four variable fields are unsigned decimal integers.  The round-trip
times given later in the output need to be divided by u to obtain
seconds.  The values c and s are specified with the -c and -s options,
and n is deduced from the input.

The next n output lines are a legend that associates small ID numbers
(in the range 0 to n-1) with the servers.  These lines have the form:

node <id> <IPaddr> <domain>

The ID numbers are assigned in the order they appear on the input.

The next group of lines, up to n*(n-1) of them, each contain a
round-trip-time measurement between an ordered pair of distinct servers.
These lines have the form:

to <id1> via <id2> rtt <rtt> at <timestamp> long <sample>... short <sample>...

The variable fields are decimal integers.  Each line corresponds to a
different (id1,id2) ordered pair.  The lines appear in the order the
measurements were taken (so the timestamp field is nondecreasing), which
is a random ordering of the (id1,id2) pairs (to avoid any systematic
bias).

The rtt is the estimated round trip between servers id1 and id2.  It is
computed as the minimum of the long samples minus the minimum of the
short samples (which means it can be inferred from the rest of the line,
but it is included for convenience).

The timestamp is the Unix time (as returned by the time() function) at
which the measurement was made.

The short samples are c individual round-trip time measurements between
the local host and server id2.  The long samples are c individual
round-trip time measurements between the local host and server id1
via server id2; that is, rtt(local,id2) + rtt(id2,id1).  Henceforth
servers id1 and id2 will be called the to-server and the via-server.
The short and long measurements are taken alternately (interleaved) and
back-to-back (at about the same time).  Just before these measurements
are taken, the DNS cache of the via-server is primed by taking a long
sample and discarding it.  This priming probe is retried until it
succeeds or has failed c times.  If the priming fails then all the long
samples automatically fail; they are not even attempted.

This batch of probes (priming, long, short) is discarded and retried
(repeatedly) in the following circumstance:  If the last long sample and
the last short sample both fail, then the network is tested.  If the
network is down then the batch is discarded, the program waits until the
network comes back up, then retries the batch (starting with the priming
probes).  The network test doesn't work well if there are less than
three unpruned servers (and probably isn't wanted in that case anyway),
so in that case the network is assumed to be up.

Once a batch of probes has been performed that does not need to be
retried, the program either outputs the line or assesses a strike
against one of the servers.  If all the short samples failed, or if all
the long samples failed and the via-server refused to perform recursion
at least once, then the line is not output and the via-server gets a
strike.  Otherwise, if all the long samples failed, the line is not
output and the to-server gets a strike.  Otherwise the line is output.
Failed long and short samples are indicated by a negative value, which
is unambiguous because negative samples are impossible.  (Different
negative values might indicate different kinds of failure, but those
meanings are deliberately undocumented and subject to change.)  Note,
however, that negative rtt values are possible (when measurement noise
is greater than the round-trip time between the servers).  Failed
samples are not included in the minimum calculation.

When a strike is assessed against a server, if the number of strikes is
still less than s, then the batch of probes for that pair of servers is
deferred and tried again later.  But if the number of strikes reaches s,
then the server is pruned from the server list and never tried again.

Each strike is either a to-strike or a via-strike, depending on which
role the server was playing when the strike was assessed.  Whenever a
batch succeeds (the line is output), the to-strikes of the to-server and
the via-strikes of the via-server server are erased.

The next group of lines is a summary of all servers that were pruned.
These lines have the form:

pruned <id>

The last line merely indicates that the program finished successfully:

finished

Error/warning messages might appear on stderr.


Compilation notes:

This is ANSI C (C89) code, plus it depends on the 4.3BSD
resolver/nameserver interface and a few POSIX-2001 interfaces:
inet_ntop(), inet_pton(), gettimeofday(), and sleep().

*/


#include <assert.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <sys/time.h>


/* failf() is like printf() but outputs to stderr */
/* and exits unsuccessfully instead of returning. */

static void failf(const char *format, ...)
{
  va_list ap;

  va_start(ap,format);
  vfprintf(stderr,format,ap);
  va_end(ap);
  exit(EXIT_FAILURE);
}


/* warnf() is like printf() but uses stderr and flushes it. */

static int warnf(const char *format, ...)
{
  va_list ap;
  int r;

  va_start(ap,format);
  r = vfprintf(stderr,format,ap);
  va_end(ap);
  fflush(stderr);
  return r;
}


static void outofmem(const char *progname)
{
  failf("%s: out of memory\n", progname);
}


static void usage(const char *progname)
{
  failf("usage:\n%s [ -c <c> ]\n", progname);
}


/* rand_ulong(max) returns a uniform random */
/* unsigned long in the range [0,max].      */

static unsigned long rand_ulong(unsigned long max)
{
  unsigned long scale, q, maxq, maxr, r;

  if (max == 0) return 0;
  if (max == RAND_MAX) return rand();

  if (max < RAND_MAX) {
    scale = ( (unsigned long) RAND_MAX
      #if RAND_MAX < ULONG_MAX
        + 1ul
      #endif
      ) / (max + 1);

    /* The numerator is RAND_MAX + 1 if that can be represented   */
    /* in an unsigned long, otherwise it's RAND_MAX.  The smaller */
    /* numerator might reduce efficiency, but it will still give  */
    /* correct results.  The denominator max + 1 cannot overflow  */
    /* because max < RAND_MAX and RAND_MAX fits in an int.        */

    do q = rand() / scale;
    while (q > max);
    return q;
  }

  /* max > RAND_MAX */

  scale = max / ((unsigned long) RAND_MAX + 1) + 1;

  /* If max > RAND_MAX, then RAND_MAX represented as an unsigned     */
  /* long can be incremented without overflow.  The scale is max + 1 */
  /* divided by RAND_MAX + 1 and rounded upward.                     */

  maxq = max / scale;
  maxr = max % scale;

  do {
    r = rand();
    q = rand_ulong(maxq);
  } while (q == maxq && r > maxr);

  return q * scale + r;
}


/* probe(addr,query,recurse) overwrites query[8..16] with a serial    */
/* number followed by a dot, sends a single probe containing the      */
/* query to addr, waits for the response, and returns the number      */
/* of milliseconds for the round-trip, or -1 if we get no response    */
/* or if the query cannot be sent, or -2 if there is a recursive      */
/* service mismatch.  The query will be recursive if recurse is       */
/* nonzero, nonrecursive otherwise, and the response will be checked  */
/* to see if got the kind of service we asked for.  Before probe()    */
/* is called, probe_init() must be called.  If the resolver state is  */
/* touched by anything other than probe(), then probe_init() must be  */
/* called again before probe() is called again.  After probe_init()   */
/* is called, the resolver state is suitable only for probe(), and    */
/* anything else wanting to use the resolver needs to call res_init() */
/* first.                                                             */

long probe(struct in_addr addr, char *query, int recurse)
{
  static unsigned long serial = 0;
  struct timeval start, finish;
  int r, qlen;
  union {
    HEADER header;
    unsigned char message[PACKETSZ];
  } answer, question;
  char textaddr[INET_ADDRSTRLEN] = {0};

  assert(serial <= 0xFFFFFFFF);
  sprintf(query + 8, "%08lx", serial++);
  query[16] = '.';
  _res.nsaddr_list[0].sin_addr = addr;
  if (recurse) _res.options |= RES_RECURSE;
  else _res.options &= ~ (unsigned long) RES_RECURSE;

#if 0  /* I don't know why res_query() doesn't work, but it doesn't... */

  r = gettimeofday(&start, 0);

  if (r == -1) {
    perror("gettimeofday");
    return -1;
  }

  r = res_query(query, C_IN, T_SOA, answer.message, PACKETSZ);

  if (r == -1) {
    inet_ntop(AF_INET, &addr, textaddr, INET_ADDRSTRLEN);
    warnf( "%lu res_query(%s, C_IN, T_SOA, answer, %d)"
           " to server %s returned -1\n",
           (unsigned long) time(0), query, PACKETSZ, textaddr );
    return -1;
  }

#else  /* ...so we'll use res_mkquery() and res_send() instead. */

  qlen = res_mkquery( QUERY, query, C_IN, T_SOA, 0, 0, 0,
                      question.message, PACKETSZ );

  if (qlen == -1) {
    warnf( "%lu res_mkquery(QUERY, %s, C_IN, T_SOA, 0, 0, 0, buf, %d)"
           " returned -1\n",
           (unsigned long) time(0), query, PACKETSZ );
    return -1;
  }

  r = gettimeofday(&start, 0);

  if (r == -1) {
    perror("gettimeofday");
    return -1;
  }

  r = res_send(question.message, qlen, answer.message, PACKETSZ);

  if (r == -1) {
    inet_ntop(AF_INET, &addr, textaddr, INET_ADDRSTRLEN);
    warnf( "%lu res_send(buf, %d, answer, %d) to server %s returned -1\n",
           (unsigned long) time(0), qlen, PACKETSZ, textaddr );
    return -1;
  }

#endif

  r = gettimeofday(&finish, 0);

  if (r == -1) {
    perror("gettimeofday");
    return -1;
  }

  if ((recurse != 0) ^ (answer.header.ra && answer.header.rd)) {
    inet_ntop(AF_INET, &addr, textaddr, INET_ADDRSTRLEN);
    warnf( ( recurse ?
             "%lu asked server %s for recursion, but didn't get it\n" :
             "%lu did not ask server %s for recursion, but got it\n" ),
           (unsigned long) time(0), textaddr );
    return -2;
  }

  return (finish.tv_sec - start.tv_sec) * 1000000 +
         finish.tv_usec - start.tv_usec;
}

void probe_init(void)
{
  res_init();
  _res.retry = 1;
  _res.nscount = 1;
}


struct server {
  char *domain;            /* Server is authoritative for domain[17..]. */
                           /* domain[0..16] are for prepending a label. */
  struct in_addr addr;     /* IP address of DNS server.                 */
  short to_strikes;        /* Number of to-strikes against server, or   */
                           /* negative if server has been pruned.       */
  short via_strikes;       /* Number of via-strikes against server, or  */
                           /* negative if server has been pruned.       */
};


/* wait_for_network(servers,n,scratch) takes an array of n servers    */
/* and randomly probes them to determine whether the network is up.   */
/* Pruned servers are not probed.  An array of n ids must be supplied */
/* as scratch space.  If the network is down, blocks until until the  */
/* network is back up.  Returns 1 if blocking occurred, 0 otherwise.  */
/* If the number of unpruned servers is less than 3, the network is   */
/* assumed to be up, and 0 is returned immediately.                   */

int wait_for_network(
  struct server *servers, unsigned int n, unsigned int *scratch )
{
  unsigned int i, nlive;
  int test1, test2, test3, blocked;

  /* Construct a list of all non-pruned server ids: */

  for (i = nlive = 0;  i < n;  ++i) {
    if (servers[i].to_strikes >= 0) scratch[nlive++] = i;
  }

  assert(nlive > 0);

  /* If all servers were already pruned, we wouldn't be      */
  /* here, because there wouldn't have been anyone to probe. */

  if (nlive < 3) return 0;

  for (blocked = 0; ; ) {

    /* Do random probes until the three     */
    /* most recent all fail or all succeed: */

    for (test1 = -1, test2 = -2, test3 = -3;
         test1 != test2 || test2 != test3;
         test1 = test2, test2 = test3) {
      i = scratch[rand_ulong(nlive - 1)];
      test3 = -1 != probe(servers[i].addr, servers[i].domain, 0);
    }

    if (test3) break;  /* network is up */

    if (!blocked) {
      blocked = 1;
      warnf("%lu network is down\n", time(0));
    }

    sleep(10);
  }

  if (blocked) warnf("%lu network is back up\n", time(0));
  return blocked;
}


struct pair {
  unsigned short to, via;
};


int main(int argc, char *argv[])
{
  time_t now;
  struct timeval tv;
  int r, in_task_list, priming_failed, recursion_refused, strikes;
  unsigned int c, s, n, i, j, to_id, via_id, *scratch_ids;
  size_t size;
  struct server *servers, tmpserver;
  char **argp, textaddr[INET_ADDRSTRLEN < 17 ? 17 : INET_ADDRSTRLEN],
    domain[240], *long_query, *short_query, *tmp;
  const char *progname = argv[0], *rp;
  long tmpl;
  unsigned long numpairs, ii, jj, randtag, serial;
  long sample, *long_samples, *short_samples, min_long, min_short, rtt;
  struct pair *pairlist, tmppair;
  struct in_addr addr;

  /* Initialization: */

  c = 4;  /* default */
  s = 1;  /* default */
  r = gettimeofday(&tv, 0);

  if (r == -1) {
    perror("gettimeofday");
    failf("%s: cannot get current time\n", progname);
  }

  srand(tv.tv_sec ^ tv.tv_usec);
  randtag = rand_ulong(0xFFFFFFFF);
  serial = 0;
  probe_init();

  /* Parse arguments: */

  for (argp = argv + 1;  *argp != 0;  ++argp) {
    if (0 == strcmp(*argp, "-c")) {
      ++argp;
      if (*argp == 0 || **argp == '\0') usage(progname);
      tmpl = strtol(*argp, &tmp, 0);
      if (*tmp != '\0') usage(progname);
      if (tmpl < 1) failf("%s: <c> must be positive\n", progname);
      c = tmpl;
    }
    else if (0 == strcmp(*argp, "-s")) {
      ++argp;
      if (*argp == 0 || **argp == '\0') usage(progname);
      tmpl = strtol(*argp, &tmp, 0);
      if (*tmp != '\0') usage(progname);
      if (tmpl < 1) failf("%s: <s> must be positive\n", progname);
      s = tmpl;
    }
    else usage(progname);
  }

  /* Read the input: */

  servers = 0;
  n = 0;
  pairlist = 0;
  numpairs = 0;
  tmpserver.to_strikes = tmpserver.via_strikes = 0;
  in_task_list = 0;

  for (;;) {
    r = scanf("%16s%239s", textaddr, domain);

    if (r == 0 || r == EOF) {
      if (feof(stdin)) break;
      failf("%s: input failure\n", progname);
    }

    if (r == 1) failf("%s: odd number of words in input\n", progname);

    if (0 == strcmp(textaddr, "pair") && 0 == strcmp(domain, "list")) {
      in_task_list = 1;
      continue;
    }

    if (strlen(textaddr) > 15) failf(
      "%s: IP address exceeded 15 characters\n", progname );
    size = strlen(domain) + 1;
    if (size > 239) failf("%s: domain exceeded 238 characters\n", progname);

    if (in_task_list) {
      tmpl = strtol(textaddr, &tmp, 0);
      if (*tmp != '\0') failf(
        "%s: non-integer <id1> = %s\n", progname, textaddr );
      if (tmpl < 0 || tmpl >= n) failf(
        "%s: <id1> = %ld, out of range\n", progname, tmpl );
      to_id = tmpl;
      tmpl = strtol(domain, &tmp, 0);
      if (*tmp != '\0') failf(
        "%s: non-integer <id2> = %s\n", progname, domain );
      if (tmpl < 0 || tmpl >= n) failf(
        "%s: <id2> = %ld, out of range\n", progname, tmpl );
      via_id = tmpl;
      if (to_id == via_id) failf("%s: <id1> = <id2> = %d\n", progname, to_id);
      ++numpairs;
      pairlist = realloc(pairlist, numpairs * sizeof (struct pair));
      if (pairlist == 0) outofmem(progname);
      pairlist[numpairs-1].to = to_id;
      pairlist[numpairs-1].via = via_id;
    }
    else {
      r = inet_pton(AF_INET, textaddr, &tmpserver.addr);
      if (r != 1) failf("%s: invalid address: %s\n", progname, textaddr);
      tmp = malloc(size + 17);
      if (tmp == 0) outofmem(progname);
      memcpy(tmp + 17, domain, size);
      sprintf(tmp, "%08lx", randtag);
      ++n;
      servers = realloc(servers, n * sizeof (struct server));
      if (servers == 0) outofmem(progname);
      servers[n-1] = tmpserver;
      servers[n-1].domain = tmp;
    }
  }

  /* Output the parameters: */

  printf( "units_per_second 1000000\n"
          "redundancy %d\n"
          "strikes_per_out %d\n"
          "nodes %d\n",
          c, s, n );

  /* Output the legend: */

  for (i = 0;  i < n;  ++i) {
    rp = inet_ntop(AF_INET, &servers[i].addr, textaddr, INET_ADDRSTRLEN);

    if (rp == 0) {
      perror("inet_ntop");
      failf("%s: inet_ntop failed\n", progname);
    }

    printf("node %d %s %s\n", i, textaddr, servers[i].domain + 17);
  }

  /* Create a shuffled list of pairs: */

  if (!in_task_list) {
    numpairs = (unsigned long) n * (n-1);
    pairlist = malloc(numpairs * sizeof (struct pair));
    if (pairlist == 0) outofmem(progname);
    ii = 0;

    for (i = 0;  i < n;  ++i) {
      for (j = 0;  j < n;  ++j) {
        if (i == j) continue;
        pairlist[ii].to = i;
        pairlist[ii].via = j;
        ++ii;
      }
    }
  }

  for (ii = 0;  ii < numpairs;  ++ii) {
    jj = ii + rand_ulong(numpairs - ii - 1);
    tmppair = pairlist[ii];
    pairlist[ii] = pairlist[jj];
    pairlist[jj] = tmppair;
  }

  /* Allocate scratch space: */

  scratch_ids = malloc(n * sizeof (unsigned int));
  long_samples = malloc(c * sizeof (int));
  short_samples = malloc(c * sizeof (int));
  if ( scratch_ids == 0 ||
       long_samples == 0 || short_samples == 0 ) outofmem(progname);

  /* Perform the measurements and output each line as we go. */

  while (numpairs > 0) {
    for (ii = 0;  ii < numpairs; ) {
      to_id = pairlist[ii].to;
      via_id = pairlist[ii].via;

      if (servers[to_id].to_strikes < 0 || servers[via_id].to_strikes < 0) {
        pairlist[ii] = pairlist[--numpairs];
        continue;
      }

      addr = servers[via_id].addr;
      long_query = servers[to_id].domain;
      short_query = servers[via_id].domain;

      for (;;) {
        recursion_refused = 0;

        /* Prime the DNS cache of server via_id: */

        for (i = 0;  i < c;  ++i) {
          sample = probe(addr, long_query, 1);

          if (sample == -2) recursion_refused = 1;
          else if (sample >= 0) break;
        }

        priming_failed = sample < 0;
        for (i = 0;  i < c;  ++i) short_samples[i] = long_samples[i] = -3;
        min_long = min_short = -4;

        for (i = 0;  i < c;  ++i) {
          sample = short_samples[i] = probe(addr, short_query, 0);
          if ( sample >= 0 &&
               (min_short < 0 || sample < min_short) ) min_short = sample;

          if (!priming_failed) {
            sample = long_samples[i] = probe(addr, long_query, 1);
            if (sample == -2) recursion_refused = 1;
            if ( sample >= 0 &&
                 (min_long < 0 || sample < min_long) ) min_long = sample;
          }
        }

        if ( short_samples[c-1] != -1 ||
             long_samples[c-1] != -1 ||
             !wait_for_network(servers, n, scratch_ids) ) break;

        /* If the last long and short probes both failed to respond, */
        /* then we test the network, and if it was down, we wait     */
        /* until it comes back up and retry.                         */

      }

      if (min_short < 0 || (min_long < 0 && recursion_refused)) {
        strikes = servers[via_id].to_strikes + ++servers[via_id].via_strikes;
        warnf( "%lu server %u not %s, now has %d strikes\n",
               (unsigned long) time(0), via_id,
               min_short < 0 ? "responding" : "recursing", strikes );

        if (strikes < s) ++ii;
        else {
          warnf("%lu pruning server %u\n", (unsigned long) time(0), via_id);
          servers[via_id].to_strikes = servers[via_id].via_strikes = -1;
        }

        continue;
      }

      if (min_long < 0) {
        strikes = ++servers[to_id].to_strikes + servers[to_id].via_strikes;
        warnf( "%lu server %u not responding, now has %hd strikes\n",
               (unsigned long) time(0), to_id, strikes );

        if (strikes < s) ++ii;
        else {
          warnf("%lu pruning server %u\n", (unsigned long) time(0), to_id);
          servers[to_id].to_strikes = servers[to_id].via_strikes = -1;
        }

        continue;
      }

      servers[via_id].via_strikes = 0;
      servers[to_id].to_strikes = 0;
      rtt = min_long - min_short;
      now = time(0);
      printf("to %d via %d rtt %ld ", to_id, via_id, rtt);
      if (now == -1) printf("-1");
      else printf("%lu", (unsigned long) now);
      printf(" long");
      for (i = 0;  i < c;  ++i) printf(" %ld", long_samples[i]);
      printf(" short");
      for (i = 0;  i < c;  ++i) printf(" %ld", short_samples[i]);
      printf("\n");
      pairlist[ii] = pairlist[--numpairs];
    }
  }

  /* Output the pruning summary: */

  for (i = 0;  i < n;  ++i) {
    if (servers[i].to_strikes < 0) printf("pruned %d\n", i);
  }

  printf("finished\n");
  return EXIT_SUCCESS;
}
